#
# See the file "LICENSE" for the full license governing this code.
#

# @summary Retrieves a control map based on module name, product version, and documentation version.
# This Puppet function supports both hashed and non-hashed control map structures.
Puppet::Functions.create_function('easy_type::get_control_map') do
  # Define the function signature
  dispatch :get_control_map do
    param 'String[1]', :module_name
    param 'String[1]', :product_version
    param 'String[1]', :doc_version
  end

  # Retrieves the control map based on the given parameters
  #
  # This method is the main entry point for the function. It determines whether
  # to use hashed values or not, and then calls the appropriate method to
  # retrieve the control map.
  #
  # @param module_name [String] The name of the module
  # @param product_version [String] The version of the product
  # @param doc_version [String] The version of the documentation
  # @return [Hash] The control map for the specified versions
  # @raise [ArgumentError] If any of the input parameters are empty or nil
  # @raise [KeyError] If the specified versions are not found in the control map
  def get_control_map(module_name, product_version, doc_version)
    @module_name = module_name
    @product_version = product_version
    @doc_version = doc_version

    if use_hashed_values?
      get_hashed_control_map
    else
      fetch_control_map
    end
  end

  private

  # Determines whether to use hashed values
  #
  # This method checks the module's configuration to see if hashed values
  # should be used for the control map.
  #
  # @return [Boolean] True if hashed values should be used, false otherwise
  def use_hashed_values?
    lookup("#{@module_name}::use_hashed_values", data_type('Boolean'), 'first', false)
  end

  # Fetches the control map from the module's configuration
  #
  # This method retrieves the entire control map from the module's
  # configuration using the Puppet lookup function. For every product and document version
  # there is a separate file containing the control map. The hiera.yml includes all
  # of these files. We use the deep merge to get all of these files into the final hash.
  #
  # @return [Hash] The complete control map
  def fetch_control_map
    lookup("#{@module_name}::control_map", data_type('Hash'), 'deep', {})
  end

  # Retrieves the hashed control map for specific versions
  #
  # This method is used when hashed values are enabled. It fetches the
  # control map and then retrieves the specific map for the given
  # product and documentation versions.
  #
  # @return [Hash] The control map for the specified versions
  # @raise [KeyError] If the specified versions are not found in the control map
  def get_hashed_control_map
    control_map = fetch_control_map
    product_map = control_map.fetch(@product_version) do
      raise KeyError, "#{@product_version} not found in #{@module_name}::control_map"
    end
    product_map.fetch(@doc_version) do
      raise KeyError, "#{@doc_version} not found in #{@module_name}::control_map[#{@product_version}]"
    end
  end

  # Parses a string into a Puppet data type
  #
  # This helper method converts a string representation of a data type
  # into an actual Puppet data type object.
  #
  # @param string [String] The string representation of the data type
  # @return [Puppet::Pops::Types::PAnyType] The parsed Puppet data type
  def data_type(string)
    parser = Puppet::Pops::Types::TypeParser.singleton
    parser.parse(string)
  end

  # Helper method for the Puppet lookup function
  #
  # This method wraps the Puppet lookup function to simplify its usage.
  #
  # @param key [String] The key to look up
  # @param type [Puppet::Pops::Types::PAnyType] The expected data type
  # @param merge [String] The merge strategy
  # @param default [Object] The default value if the key is not found
  # @return [Object] The value retrieved by the lookup function
  def lookup(key, type, merge, default)
    call_function('lookup', key, type, merge, default)
  end

end
