#
# See the file "LICENSE" for the full license governing this code.
#

# This Puppet function retrieves control data from a specified profile.
# It supports both hashed and non-hashed profile data structures.
Puppet::Functions.create_function('easy_type::get_controls_in_profile') do
  # Define the function signature
  # @param module_name [String] The name of the module
  # @param product_version [String] The version of the product
  # @param doc_version [String] The version of the documentation
  # @param profile [String] The profile name
  dispatch :get_controls_in_profile do
    param 'String[1]', :module_name
    param 'String[1]', :product_version
    param 'String[1]', :doc_version
    param 'String[1]', :profile
  end

  # Retrieves control data from the specified profile.
  # Depending on the configuration, it fetches either hashed or non-hashed profile data.
  #
  # @param module_name [String] The name of the module
  # @param product_version [String] The version of the product
  # @param doc_version [String] The version of the documentation
  # @param profile [String] The profile name
  # @return [Hash] The control data from the profile
  def get_controls_in_profile(module_name, product_version, doc_version, profile)
    @module_name = module_name
    @profile = profile
    @product_version = product_version
    @doc_version = doc_version

    if use_hashed_values?
      get_hashed_profile_data
    else
      fetch_old_profile_data
    end
  end

  private

  # Determines if hashed values should be used for profile data.
  #
  # @return [Boolean] True if hashed values should be used, false otherwise
  def use_hashed_values?
    lookup("#{@module_name}::use_hashed_values", data_type('Boolean'), 'first', false)
  end

  # Fetches the profile data from the module.
  #
  # @return [Hash] The profile data
  def fetch_profile_data
    lookup("#{@module_name}::profile_data", data_type('Hash'), 'deep', {})
  end

  # Fetches the old profile data structure.
  #
  # @return [Hash] The control data from the old profile structure
  # @raise [KeyError] If the 'controls' key is not found in the profile data
  def fetch_old_profile_data
    hiera_data = lookup("#{@module_name}::profile::#{@profile}", data_type('Hash'), 'first', {})
    hiera_data.fetch('controls') do
      raise KeyError, "key controls not found in #{@module_name}::profile::#{@profile}"
    end
  end

  # Fetches the hashed profile data structure.
  #
  # @return [Hash] The control data from the hashed profile structure
  # @raise [KeyError] If any of the required keys are not found in the profile data
  def get_hashed_profile_data
    product_data = fetch_profile_data.fetch(@product_version) do
      raise KeyError, "#{@product_version} not found in #{@module_name}::profile_data"
    end
    version_specific_data = product_data.fetch(@doc_version) do
      raise KeyError, "#{@doc_version} not found in #{@module_name}::profile_data[#{@product_version}]"
    end
    profile_data = version_specific_data.fetch(@profile) do
      raise KeyError, "#{@profile} not found in #{@module_name}::profile_data[#{@product_version}][#{@doc_version}]"
    end
    profile_data.fetch('controls') do
      raise KeyError, "key controls not found in #{@module_name}::profile_data[#{@product_version}][#{@doc_version}][#{@profile}]"
    end
  end

  # Parses a string into a Puppet data type.
  #
  # This helper method converts a string representation of a data type
  # into an actual Puppet data type object.
  #
  # @param string [String] The string representation of the data type
  # @return [Puppet::Pops::Types::PAnyType] The parsed Puppet data type
  def data_type(string)
    parser = Puppet::Pops::Types::TypeParser.singleton
    parser.parse(string)
  end

  # Helper method to perform lookup calls.
  #
  # @param key [String] The key to lookup
  # @param type [Puppet::Pops::Types::PAnyType] The expected data type
  # @param merge [String] The merge strategy
  # @param default [Object] The default value if the key is not found
  # @return [Object] The result of the lookup
  def lookup(key, type, merge, default)
    call_function('lookup', key, type, merge, default)
  end

end
