#
# See the file "LICENSE" for the full license governing this code.
#
# frozen_string_literal: true
#
# This is a hiera lookup backend that you can use to fetch data based on locale puppet
# variables without hiera giving you a warning when the variable doesn't exist. It can only 
# be used in the context of a module at this point in time.
#
# You have to specify the `file_name` option in the `hiera.yaml`. Puppet variables
# can be used in this entry just like in regular puppet interpollation. Here is an example:
#
# ```yaml
#    - name: "Control data"
#    data_hash:  easy_type::yaml_data_without_warnings
#    options:
#    file_name: "benchmarks/cis/${db_version}/${doc_version}/control_data.yaml"
# ```
#
require 'yaml'
require 'thread'
require 'concurrent-ruby'

Puppet::Functions.create_function('easy_type::yaml_data_without_warnings') do
  # Mutex for synchronizing file I/O and cache access
  @@file_mutex = Mutex.new

  
  # Mutex for synchronizing data merging operations
  @@data_mutex = Mutex.new

  dispatch :yaml_data_without_warnings do
    param 'Hash', :options
    param 'Any', :context
  end

  def yaml_data_without_warnings(options, context)
    # Check if we are running in the module context
    if context.module_name.to_s == ''
      context.explain { "Running outside of module scope. skipping lookup." }
      return context.not_found
    end

    # Process options
    file_name = options['file_name']
    raise Puppet::Error.new("the 'easy_type::yaml_data_without_warnings' backend needs a file_name option field") if file_name.nil?

    context.explain { "File specified in hiera.yaml is '#{file_name}'" }

    # Fetch and verify variables
    variable_values = {}
    variable_names = file_name.scan(/\$\{(.+?)\}/)&.flatten
    variable_names.each do |variable_name|
      variable_values[variable_name] = fetch_variable(context, variable_name)
      if variable_values[variable_name].nil? || (variable_values[variable_name].is_a?(Array) && variable_values[variable_name].empty?)
        context.explain { "Variable '#{variable_name}' not set, or empty; skipping this lookup." }
        return context.not_found
      end
    end

    # Create possible combinations for file names
    file_names = []
    all_combinations(variable_values.values).each do |combination|
      result = file_name.dup
      variable_names.each_with_index { |name, index| result.gsub!("${#{name}}", combination[index]) }
      file_names << result
    end

    # Thread-safe data storage using Concurrent::Hash
    data = Concurrent::Hash.new

    # Process each file name
    file_names.each do |file_name|
      context.explain { "File name after variable substitution is '#{file_name}'" }

      data_dir = File.expand_path(File.dirname(__FILE__) + "/../../../../../#{context.module_name}/data")
      path = "#{data_dir}/#{file_name}"

      # Synchronize file existence check and file read operations
      @@file_mutex.synchronize do
        unless File.exist?(path)
          context.explain { "File '#{path}' not found, skipping this lookup." }
          return context.not_found
        end

        # Fetch and merge file data in a thread-safe way
        context.cached_file_data(path) do |content|
          begin
            new_data = Puppet::Util::Yaml.safe_load(content, [Symbol], path)

            if new_data.is_a?(Hash)
              # Synchronize the merge operation to ensure thread safety
              @@data_mutex.synchronize do
                data.merge!(Puppet::Pops::Lookup::HieraConfig.symkeys_to_string(new_data))
              end
            else
              msg = _("%{path}: file does not contain a valid yaml hash" % { path: path })
              raise Puppet::DataBinding::LookupError, msg if Puppet[:strict] == :error && data != false
              Puppet.warning(msg)
              {}
            end
            data.to_h
          rescue Puppet::Util::Yaml::YamlLoadError => ex
            raise Puppet::DataBinding::LookupError, _("Unable to parse %{message}") % { message: ex.message }
          end
        end
      end
    end

    data.to_h
  end

  # Helper method to create combinations for variable substitution
  def all_combinations(array)
    result = []
    length = array.map { |e| Array(e).length }.inject(:*)
    array.each_with_index do |value, index|
      if value.is_a?(Array)
        (0...length).each { |i| result[i] ||= []; result[i][index] = value[i % value.length] }
      else
        (0...length).each { |i| result[i] ||= []; result[i][index] = value }
      end
    end
    result
  end

  # Fetch the variable from the context
  def fetch_variable(context, name)
    return nil unless context.invocation.scope.include?(name)
    context.invocation.scope[name]
  end

  # Error message for missing path
  def missing_path(options, context)
    "One of 'path', 'paths', 'glob', 'globs' or 'mapped_paths' must be declared in hiera.yaml when using this data_hash function."
  end
end
