# frozen_string_literal: true

#
# See the file "LICENSE" for the full license governing this code.
#
require 'easy_type'
require 'puppet/provider/exec/windows'

Puppet::Type.type(:validation).provide :regval, :parent => Puppet::Type::Exec::ProviderWindows do
  confine 'os.name' => :windows
  include EasyType::Provider

  @output = ''

  desc 'Execute a validation on the Windows registry keys'

  mk_resource_methods

  def validate_run?
    hive(resource[:check]).class == Win32::Registry
  end

  def hive(check)
    hivekey = check.split('\\').first
    case hivekey
    when 'HKCR'
      Win32::Registry::HKEY_CLASSES_ROOT
    when 'HKCU'
      Win32::Registry::HKEY_CURRENT_USER
    when 'HKLM'
      Win32::Registry::HKEY_LOCAL_MACHINE
    when 'HKU'
      Win32::Registry::HKEY_USERS
    when 'HKCC'
      Win32::Registry::HKEY_CURRENT_CONFIG
    else
      begin
        Object.const_get("Win32::Registry::#{hivekey}")
      rescue
        Puppet.debug "Unable to load registry values for hive key #{hivekey}"
        raise Puppet::Error.new "Unable to load registry values for hive key #{hivekey}"
      end
    end
  end

  def execute_check
    @output = run(resource.check)
  end

  def report
    if resource.fail_message != :absent
      message = resource[:fail_message] % {output: @output}
    else
      message = "Validation #{name}  failed; expected: '#{resource.expected_output}' got : '\n#{@output}'"
    end
    message = "#{resource.path}: #{message}"
    case resource.report_as
    when /warning/i
      Puppet.warning message
    when /error/i
      raise Puppet::Error, message
    when /command/i
      # Issue a waning and execute the fail_command
      Puppet.warning message
      final_command = resource.fail_command % {output: @output}
      run(final_command, true, resource.parameters[:fail_command].sensitive).first.to_s
    when /none/i
      Puppet.debug message
    else
      raise "#{resource.report_as} is not a valid value"
    end
  end

  def run(key)
    hive_key = key.split('\\').first
    if hive_key == 'HKU' || hive_key == 'HKEY_USERS' && key.include?('[USER SID]')
      # Usecase for 'wildcards' ([USER SID]) in registry key
      output = ''

      regpath, regkey_name = key.split(':')
      parent = regpath.split('\\').drop(2).join('\\')

      # Get the SIDs from all the users
      profiles = hive('HKLM').open('SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList')
      profiles.keys.each do |sid|
        next unless hive('HKLM').open("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList\\#{sid}")['ProfileImagePath'].include?('\\Users\\')
        next if hive('HKLM').open("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList\\#{sid}")['ProfileImagePath'].include?('\\Administrator')

        begin
          contents = hive(resource[:check]).open("#{sid}\\#{parent}")
          Puppet.debug "#{key.gsub('[USER SID]', sid)}: #{contents[regkey_name]}"
          output += "#{contents[regkey_name]}"
        rescue
          Puppet.debug "Key #{regkey_name} does not exist in registry path #{key.gsub('[USER SID]', sid)}"
          output += '_'
        end
      end
      output = output.split('').uniq.join
    else
      regpath, regkey_name = key.split(':')
      parent = regpath.split('\\').drop(1).join('\\')

      debug "Checking registry key #{regpath} for value of #{regkey_name}"

      begin
        contents = hive(resource[:check]).open(parent)
        Puppet.debug "#{key}: #{contents[regkey_name]}"
        output = contents[regkey_name]
      rescue
        Puppet.debug "Key #{regkey_name} does not exist in registry path #{regpath}"
        output = 'Key not found'
      end
    end

    case output
    when Integer
      return output.to_s
    when Array
      return output.join(',')
    else
      return output
    end
  end
end
